"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                           MATRICES                                          ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

   In mathematics a matrix is a rectangular array of numbers. It can be multiplied by and added to other
matrices, according to certain rules. But in computer science a matrix is usually a rectangular table of
numbers represented as a doubly subscripted array of any data type. In Python a single matrix is allowed to
have cells of different data types. In fact, the length of the rows (or columns) need not be the same.

#---How to create a small matrix, by hand.
    M = [[0,1,2,3,4,6],
         [7,8,9,9,9,7],
         [6,5,0,4,3,2],
         [1,0,1,2,3,4],
         [5,6,7,8,9,8],
         [7,6,5,4,6,3],]

#---How to reference an element:
    print(M[2][4]) #  <-- Note: M[2][4] = 3 = M[row][col]

#---How to pretty-print a matrix.
    print ('---MATRIX:')
    for row in M:
       [print ('%11.2f'% x, end = '') for x in row]
       print()
    print ('   ==============================')

# output:
#---MATRIX:
#       0.00       1.00       2.00       3.00       4.00       6.00
#       7.00       8.00       9.00       9.00       9.00       7.00
#       6.00       5.00       0.00       4.00       3.00       2.00
#       1.00       0.00       1.00       2.00       3.00       4.00
#       5.00       6.00       7.00       8.00       9.00       8.00
#       7.00       6.00       5.00       4.00       6.00       3.00

#---How to build a matrix with list comprehensions.
    ROW = 3
    COL = 5
    M = [[c for r in range(ROW)]  for c in range(COL)]
    print (M)
    # output: [[0, 0, 0], [1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]]

#---Again, we reference by row first, and then column second.
    M[4][2] += 100

#--How to adjust every element in the matrix (by adding 1 in this case)
    for n in list(range(ROW*COL)):
        r = n  % COL
        c = n // COL # <-- Yes, use COL again. In Python 2.x, use "n / COL" instead.
        M[r][c] += 1
    print (M)
    # output: [[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4], [5, 5, 105]]
    # (Thanks to Kalki Sindhwani in 2011 for this clever code above.)

#--How to transpose a matrix.
    M = list(zip(*M))
    print(M)    # [(1, 2, 3, 4, 5), (1, 2, 3, 4, 5), (1, 2, 3, 4, 105)] <-- the transpose = list of TUPLES
    M =[list(x) for x in M]
    print(M)    # [[1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 2, 3, 4, 105]] <-- the transpose = list of LISTS


#--A matrix (set of lists) that indicate the cities that one can travel to is called an adjacency matrix or a
#  set of adjacency lists. A "directed" graph has one-way highways.]



    A,B,C,D,E,F = range(6)   # <-- Six assignments are in one line (efficient!).
    inf = _ = float('inf')   # <-- We define both the underscore '_' and inf, as bigger than any integer.

#        A B C D E F
    M = [[0,7,_,3,4,_],# A    <-- Note: M[1][4] = 7 = M[row][col]
        [_,0,_,2,7,_],# B
        [5,_,0,_,_,3],# C    <-- Using '_' makes the matrix easy to read.
        [_,_,1,0,_,9],# D
        [4,_,6,_,0,_],# E
        [1,_,_,_,6,0]]# F

#  1. Print the distance from city A to city B.
    print(M[A][B])                               # output: 7 (kilometers)

#  2. There is a connection from city A (row) to city B (col)
    print(M[A][B] < inf)                         # output: True

#  3. Produce the set of the distances to other cities from city A.
    print(set(x for x in M[A] if x < inf) - {0}) # output: {3, 4, 7}

#  4. Print the number of cities that A connects to.
    print( sum(1 for x in M[A] if x < inf) - 1 ) # output: 3

#  5. Produce a list of city number-names that city A connects to.
    nums = [n for n in range(6) if M[A][n] < inf and 0 < M[A][n]]
    print(nums)                                  # output: [1, 3, 4]

#  6. Translate the city numbers into characters ('A' to 'F').
    table = {0:'A', 1:'B', 2:'C', 3:'D', 4:'E', 5:'F'}
    print([table[city] for city in nums]) # output: ['B', 'D', 'E']
